import os
import numpy as np
import matplotlib.pyplot as plt 
from gasp import dataloader, get_project_path

def load_dataset(url, foldername, files = None, path = get_project_path()):
    dataloader.download_data(url, foldername, path)
    filepath =  os.path.join(path, 'data', foldername, '')       
    if files is None:
        files = os.listdir(filepath)
        files.sort()
    data_list = [dataloader.read_rawdata(os.path.join(filepath, file))['data'] for file in files]
    M = np.stack(data_list, axis=-1)
    return M

def load_dataset0(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1ttvKQUAPjRJdtocAIJzUv1_7n3kBEOSI/view?usp=share_link'
    dataloader.download_data(url, '20190401_GASP_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20190401_GASP_PHANTOM', '')   
    files = ['meas_MID48_TRUFI_NBPM_2019_02_27_FID41503.dat',
            'meas_MID49_TRUFI_NBPM_2019_02_27_FID41504.dat',
            'meas_MID50_TRUFI_NBPM_2019_02_27_FID41505.dat']

    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset1(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1ttvKQUAPjRJdtocAIJzUv1_7n3kBEOSI/view?usp=share_link'
    dataloader.download_data(url, '20190401_GASP_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20190401_GASP_PHANTOM', '')   
    files = ['meas_MID54_TRUFI_NBPM_2019_02_27_FID41509.dat',
            'meas_MID55_TRUFI_NBPM_2019_02_27_FID41510.dat',
            'meas_MID56_TRUFI_NBPM_2019_02_27_FID41511.dat']

    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset2(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1l-JqXUnn7WVubMRaSI1uVsenuQg1MUeS/view?usp=share_link'
    dataloader.download_data(url, '20190507_GASP_LONG_TR_WATER', path)
    filepath =  os.path.join(path, 'data', '20190507_GASP_LONG_TR_WATER', '')   
    files = ['meas_MID12_TRUFI_TE12_FID42712.dat',
            'meas_MID13_TRUFI_TE24_FID42713.dat',
            'meas_MID14_TRUFI_TE48_FID42714.dat']

    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset3(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1k4NIhvuCBZKDNKcF4aFhtXEL2JbKw8VU/view?usp=sharing'
    dataloader.download_data(url,'20190506_GASP_CRISCO_WATER_PHANTOMS', path)
    filepath =  os.path.join(path, 'data', '20190506_GASP_CRISCO_WATER_PHANTOMS', '') 
    files = ['meas_MID22_TRUFI_TE12_FID42700.dat',
             'meas_MID24_TRUFI_TE24_FID42702.dat',
             'meas_MID25_TRUFI_TE48_FID42703.dat']

    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset4(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1yxmriHAoNubNtyZDPak-TmM29tBHDmFv/view?usp=sharing'
    dataloader.download_data(url,'20190508_GASP_WATER_FAT_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20190508_GASP_WATER_FAT_PHANTOM', '') 
    files = ['meas_MID28_TRUFI_TE3_FID42728.dat',
             'meas_MID31_TRUFI_TE6_FID42731.dat',
             'meas_MID33_TRUFI_TE12_FID42733.dat']
    
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset5(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1O8xm9yWk-3vA8H90d3bVer7Ec8u143G-/view?usp=sharing'
    dataloader.download_data(url, '20190812_GASP_INVIVO_Sag_Knee', path)
    filepath =  os.path.join(path, 'data', '20190812_GASP_INVIVO_Sag_Knee', '') 
    files = ['meas_MID131_TRUFI_TE3_FID48578.dat', 
             'meas_MID132_TRUFI_TE6_FID48579.dat',
             'meas_MID133_TRUFI_TE12_FID48580.dat']
    
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset6a(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/18NV--KkY9QmXm9OVSL73Lvm8Iqsw7ALH/view?usp=sharing'
    dataloader.download_data(url, '20190827_GASP_INVIVO_BRAIN_HIP', path)
    filepath =  os.path.join(path, 'data', '20190827_GASP_INVIVO_BRAIN_HIP', '') 
    files = ['meas_MID299_TRUFI_TE3_FID49324.dat',
             'meas_MID300_TRUFI_TE6_FID49325.dat',
             'meas_MID301_TRUFI_TE12_FID49326.dat']
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M
    
def load_dataset7a():
    url = 'https://drive.google.com/file/d/11szQZR8MPmT09zaM-lCSc4nUlNC4E-el/view?usp=sharing'
    foldername = '20231106_GASP_PHANTOM'
    files = ['meas_MID162_bSSFP_gasp_knee_fa90_1x1x2_2D_TR6ms_FID55595.dat',
            'meas_MID163_bSSFP_gasp_knee_fa90_1x1x2_2D_TR12ms_FID55596.dat',
            'meas_MID164_bSSFP_gasp_knee_fa90_1x1x2_2D_TR24ms_FID55597.dat']  
    return load_dataset(url, foldername, files)

def load_dataset7b():
    url = 'https://drive.google.com/file/d/11szQZR8MPmT09zaM-lCSc4nUlNC4E-el/view?usp=sharing'
    foldername = '20231106_GASP_PHANTOM'
    files = ['meas_MID165_bSSFP_gasp_knee_fa20_1x1x2_2D_TR6ms_FID55598.dat',
            'meas_MID166_bSSFP_gasp_knee_fa20_1x1x2_2D_TR12ms_FID55599.dat',
            'meas_MID167_bSSFP_gasp_knee_fa20_1x1x2_2D_TR24ms_FID55600.dat']  
    return load_dataset(url, foldername, files)

def load_dataset8a(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1pbn2gsxe-PjvW9vSIE6pnfbCUVWrUGbe/view?usp=sharing'
    dataloader.download_data(url, '20231106_GASP_KNEE', path)
    filepath =  os.path.join(path, 'data', '20231106_GASP_KNEE', '') 
    files = ['meas_MID123_bSSFP_gasp_knee_faMax_1x1x2_2D_TR6ms_FID55556.dat',
             'meas_MID124_bSSFP_gasp_knee_faMax_1x1x2_2D_TR12ms_FID55557.dat',
             'meas_MID125_bSSFP_gasp_knee_faMax_1x1x2_2D_TR24ms_FID55558.dat']

    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset8b(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1pbn2gsxe-PjvW9vSIE6pnfbCUVWrUGbe/view?usp=sharing'
    dataloader.download_data(url, '20231106_GASP_KNEE', path)
    filepath =  os.path.join(path, 'data', '20231106_GASP_KNEE', '') 
    files = ['meas_MID127_bSSFP_gasp_knee_fa20_1x1x2_2D_TR6ms_FID55560.dat',
             'meas_MID128_bSSFP_gasp_knee_fa20_1x1x2_2D_TR12ms_FID55561.dat',
             'meas_MID129_bSSFP_gasp_knee_fa20_1x1x2_2D_TR24ms_FID55562.dat']

    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset8c(path = os.getcwd()):
    url = 'https://drive.google.com/file/d/1pbn2gsxe-PjvW9vSIE6pnfbCUVWrUGbe/view?usp=sharing'
    dataloader.download_data(url, '20231106_GASP_KNEE', path)
    filepath =  os.path.join(path, 'data', '20231106_GASP_KNEE', '') 
    files = ['meas_MID126_DIXON_2D_3echoes_FID55559.dat']
    M = dataloader.read_rawdata(filepath + files[0])['data']
    return M

def load_dataset9a(path = os.getcwd()):
    ''' Retreives GASP Phantom data for dixon - Experiment from Dec 22, 2023 '''
    url = 'https://drive.google.com/file/d/10ZRAlIO9w5Q3EsJLHXpnIwU14HM9n12L/view?usp=sharing'
    dataloader.download_data(url, '20231222_GASP_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20231222_GASP_PHANTOM', 'dixon', '') 
    files = os.listdir(filepath)
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset9b(path = os.getcwd()):
    ''' Retreives GASP Phantom data for fa20 - Experiment from Dec 22, 2023 '''
    url = 'https://drive.google.com/file/d/10ZRAlIO9w5Q3EsJLHXpnIwU14HM9n12L/view?usp=sharing'
    dataloader.download_data(url, '20231222_GASP_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20231222_GASP_PHANTOM', 'gasp_fa20', '') 
    files = os.listdir(filepath)
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset9c(path = os.getcwd()):
    ''' Retreives GASP Phantom data for fa90 - Experiment from Dec 22, 2023 '''
    url = 'https://drive.google.com/file/d/10ZRAlIO9w5Q3EsJLHXpnIwU14HM9n12L/view?usp=sharing'
    dataloader.download_data(url, '20231222_GASP_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20231222_GASP_PHANTOM', 'gasp_fa90', '') 
    files = os.listdir(filepath)
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset10a(path = os.getcwd()):
    ''' Retreives GASP Ankle data for dixon - Experiment from Dec 22, 2023 '''
    url = 'https://drive.google.com/file/d/10hHegaWbiDYv4MsOt8nXDLpxl1b1xccE/view?usp=sharing'
    dataloader.download_data(url, '20231222_GASP_ANKLE', path)
    filepath =  os.path.join(path, 'data', '20231222_GASP_ANKLE', 'dixon', '') 
    files = os.listdir(filepath)
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset10b(path = os.getcwd()):
    ''' Retreives GASP Ankle data for fa20 - Experiment from Dec 22, 2023 '''
    url = 'https://drive.google.com/file/d/10hHegaWbiDYv4MsOt8nXDLpxl1b1xccE/view?usp=sharing'
    dataloader.download_data(url, '20231222_GASP_ANKLE', path)
    filepath =  os.path.join(path, 'data', '20231222_GASP_ANKLE', 'fa20', '') 
    files = os.listdir(filepath)
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset10c(path = os.getcwd()):
    ''' Retreives GASP Ankle data for fa90 - Experiment from Dec 22, 2023 '''
    url = 'https://drive.google.com/file/d/10hHegaWbiDYv4MsOt8nXDLpxl1b1xccE/view?usp=sharing'
    dataloader.download_data(url, '20231222_GASP_ANKLE', path)
    filepath =  os.path.join(path, 'data', '20231222_GASP_ANKLE', 'fa90', '') 
    files = os.listdir(filepath)
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset11(path = os.getcwd(), foldername='dixon'):
    ''' Retreives GASP Ankle data for fa90 - Experiment from March 12, 2024 '''
    url = 'https://drive.google.com/file/d/1M1WdParsJlWMd5es3ve_e3_lfjS9O-Bl/view?usp=sharing'
    dataloader.download_data(url, '20240312_GASP_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20240312_GASP_PHANTOM', foldername, '') 
    files = os.listdir(filepath)
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset11a(path = os.getcwd()):
    return load_dataset11(path, 'dixon')

def load_dataset11b(path = os.getcwd()):
    return load_dataset11(path, 'fa20')

def load_dataset11c(path = os.getcwd()):
    return load_dataset11(path, 'fa90')

def load_dataset12(path = os.getcwd(), foldername='dixon'):
    ''' Retreives GASP Phantom for dixon - Experiment from March 27, 2024 '''
    url = 'https://drive.google.com/file/d/1kxqMLtBhsXH0DN4UbgCYwIkjRbSM2CMA/view?usp=sharing'
    dataloader.download_data(url, '20240327_GASP_PHANTOM', path)
    filepath =  os.path.join(path, 'data', '20240327_GASP_PHANTOM', foldername, '')
    files = os.listdir(filepath)
    files.sort()
    print(f'Path: {filepath}')
    print(f'Loading files: {files}')
    M0 = dataloader.read_rawdata(filepath + files[0])['data']
    M1 = dataloader.read_rawdata(filepath + files[1])['data']
    M2 = dataloader.read_rawdata(filepath + files[2])['data']
    M = np.stack([M0,M1,M2], axis=-1)
    return M

def load_dataset12a(path = os.getcwd()):
    return load_dataset12(path, 'dixon')

def load_dataset12b(path = os.getcwd()):
    return load_dataset12(path, 'fa20')

def load_dataset12c(path = os.getcwd()):
    return load_dataset12(path, 'fa90')

def load_dataset12d(path = os.getcwd()):
    return load_dataset12(path, 'dixon2')


"""
responses.py: Standardized Filter Response Functions for Signal Processing

This module provides a collection of filter response functions commonly used in
signal processing and spectral analysis. Each function generates a response curve
with a standardized bandwidth and shift parameter. The available response types
include Square, Gaussian, Sinc, Lorentzian, and Butterworth filters.

The module also includes generic bandpass and stopband filter functions that can
use any of the implemented response types. All functions operate on a normalized
frequency range of [-1, 1] for consistency and easy comparison.

Functions:
    square, gaussian, sinc, lorentzian, butterworth: Individual filter responses
    bandpass: Generates a bandpass filter of the specified type
    stopband: Generates a stopband filter of the specified type

Usage:
    Import this module to access standardized filter response functions for
    various signal processing applications, spectral shaping, or filter design.
    For example useage, can run this module as a script to generate and plot all 
    response curves.
"""

import numpy as np
from scipy import signal

def square(width: int, bw: float, shift: float) -> np.ndarray:
    """
    Generate a square response curve with standardized bandwidth.

    Args:
        width (int): Number of points in the output array.
        bw (float): Bandwidth of the response, normalized to [0, 1].
        shift (float): Amount to shift the response, normalized to [-1, 1].

    Returns:
        np.ndarray: Square response curve.
    """
    bw = 2 * bw  # scale bandwidth to normalized frequency range
    x = np.linspace(-1, 1, width)
    y = np.zeros(width)
    y[np.abs(x) < bw/2] = 1
    return np.roll(y, int(shift * width))

def gaussian(width: int, bw: float, shift: float) -> np.ndarray:
    """
    Generate a Gaussian response curve with standardized bandwidth.

    Args:
        width (int): Number of points in the output array.
        bw (float): Bandwidth of the response, normalized to [0, 1].
        shift (float): Amount to shift the response, normalized to [-1, 1].

    Returns:
        np.ndarray: Gaussian response curve.
    """
    bw = 2 * bw  # scale bandwidth to normalized frequency range
    x = np.linspace(-1, 1, width)
    sigma = bw / 2.355  # FWHM = 2.355 * sigma for Gaussian
    y = np.exp(-(x**2) / (2 * sigma**2))
    return np.roll(y, int(shift * width))

def sinc(width: int, bw: float, shift: float) -> np.ndarray:
    """
    Generate a sinc response curve with standardized bandwidth.

    Args:
        width (int): Number of points in the output array.
        bw (float): Bandwidth of the response, normalized to [0, 1].
        shift (float): Amount to shift the response, normalized to [-1, 1].

    Returns:
        np.ndarray: Sinc response curve.
    """
    bw = 2 * bw  # scale bandwidth to normalized frequency range
    x = np.linspace(-1, 1, width)
    # The FWHM of an unscaled sinc is approximately 1.20784
    scale_factor = 1.20784 / bw
    y = np.sinc(x * scale_factor)
    return np.roll(y, int(shift * width))

def lorentzian(width: int, bw: float, shift: float) -> np.ndarray:
    """
    Generate a Lorentzian response curve with standardized bandwidth.

    Args:
        width (int): Number of points in the output array.
        bw (float): Bandwidth of the response, normalized to [0, 1].
        shift (float): Amount to shift the response, normalized to [-1, 1].

    Returns:
        np.ndarray: Lorentzian response curve.
    """
    bw = 2 * bw  # scale bandwidth to normalized frequency range
    x = np.linspace(-1, 1, width)
    gamma = bw / 2  # FWHM = 2 * gamma for Lorentzian
    y = 1 / (1 + (x / gamma)**2)
    return np.roll(y, int(shift * width))

def butterworth(width: int, bw: float, shift: float) -> np.ndarray:
    """
    Generate a Butterworth bandpass filter response with standardized bandwidth.

    Args:
        width (int): Number of points in the output array.
        bw (float): Bandwidth of the response, normalized to [0, 1].
        shift (float): Amount to shift the response, normalized to [-1, 1].

    Returns:
        np.ndarray: Butterworth bandpass filter response.
    """
    low = (1 - bw) / 2
    high = (1 + bw) / 2
    b, a = signal.butter(5, [low, high], btype='band')
    w, h = signal.freqz(b, a, worN=width)
    y = np.abs(h)
    return np.roll(y, int(shift * width))

def bandpass(width: int, bw: float, shift: float, type: str = 'butterworth') -> np.ndarray:
    """
    Generate a bandpass filter response with standardized bandwidth.

    Args:
        width (int): Number of points in the output array.
        bw (float): Bandwidth of the response, normalized to [0, 1].
        shift (float): Amount to shift the response, normalized to [-1, 1].
        type (str, optional): Type of bandpass filter. Defaults to 'butterworth'.
            Options: 'Square', 'Gaussian', 'Sinc', 'Lorentzian', 'Butterworth'.

    Returns:
        np.ndarray: Bandpass filter response of the specified type.

    Raises:
        ValueError: If an invalid filter type is specified.
    """
    responses = {
        'square': square,
        'gaussian': gaussian,
        'sinc': sinc,
        'lorentzian': lorentzian,
        'butterworth': butterworth
    }

    type = type.lower()
    if type not in responses:
        raise ValueError(f"Invalid filter type. Choose from: {', '.join(responses.keys())}")
    return responses[type](width, bw, shift)

def stopband(width: int, bw: float, shift: float, type: str = 'butterworth') -> np.ndarray:
    """
    Generate a stopband filter response with standardized bandwidth.

    Args:
        width (int): Number of points in the output array.
        bw (float): Bandwidth of the response, normalized to [0, 1].
        shift (float): Amount to shift the response, normalized to [-1, 1].
        type (str, optional): Type of stopband filter. Defaults to 'butterworth'.
            Options: 'Square', 'Gaussian', 'Sinc', 'Lorentzian', 'Butterworth'.

    Returns:
        np.ndarray: Stopband filter response of the specified type.
    """
    return 1 - bandpass(width, bw, shift, type)

if __name__ == "__main__":
    # Example usage
    
    def plot_all_responses(width: int, bw: float, shift: float):
        """ Generate and plot all response curves """

        x = np.linspace(-1, 1, width)
        responses = {
            'square': square(width, bw, shift),
            'gaussian': gaussian(width, bw, shift),
            'sinc': sinc(width, bw, shift),
            'lorentzian': lorentzian(width, bw, shift),
            'butterworth': butterworth(width, bw, shift)
        }

        import matplotlib.pyplot as plt
        plt.figure(figsize=(12, 8))
        for name, y in responses.items():
            plt.plot(x, y, label=name)

        plt.legend()
        plt.title(f'Response Functions (BW={bw}, Shift={shift})')
        plt.xlabel('Normalized Frequency')
        plt.ylabel('Amplitude')
        plt.grid(True)

        _3db = 10 ** (-3/20)
        plt.axhline(_3db, color='k', linestyle=':', label='-3 dB Line')
        plt.axhline(0.5, color='k', linestyle=':', label='-3 dB Line')
        plt.show()

    plot_all_responses(1000, 0.2, 0)


from typing import Tuple
import numpy as np
from gasp import phantom

# T1/T2 values taken from https://mri-q.com/why-is-t1--t2.html
tissue_map = {
    'none': [0, 0, 0],
    'water': [4, 2, 0],
    'white-matter': [0.6, 0.08, 0],
    'gray-matter': [0.9, 0.1, 0],
    'muscle': [0.9, .05, 0],
    'liver': [0.5, 0.04, 0],
    'fat': [0.25, 0.07, 420], # 420Hz is the average fat signal
    'tendon': [0.4, 0.005, 0],
    'proteins': [0.250, 0.001, 0]
}

def get_t1(x: int) -> Tuple[float, float]:
    x = int(x)
    keys = list(tissue_map.keys())
    return tissue_map[keys[x]][0]

def get_t2(x: int) -> Tuple[float, float]:
    x = int(x)
    keys = list(tissue_map.keys())
    return tissue_map[keys[x]][1]

def get_f0(x: int) -> Tuple[float, float]:
    x = int(x)
    keys = list(tissue_map.keys())
    return tissue_map[keys[x]][2]

def tissue_generator(fov: int=256, type: str='blocks', padding: int = 8):
    """ Generates a tissue phantom with a given shape for a number of coils.
    Args:
        fov: size of image.
        type: type of phantom
    Returns:
        tissue dict
    """
    fov = int(fov)
    keys = list(tissue_map.keys())

    if type == 'shepp_logan':
        img = phantom.shepp_logan_phantom([fov, fov])
    elif type == 'circle':
        img = phantom.circle_phantom([fov, fov])
    elif type == 'circles':
        img = phantom.circle_array_phantom([fov, fov])
    elif type == 'block':
        img = phantom.block_phantom_single(fov, padding=padding)
    elif type == 'blocks':
        img = phantom.block_phantom()
    elif type == 'line':
        img = phantom.line_phantom(fov, padding=padding)
    elif type == 'line-whitematter':
        img = phantom.line_phantom(fov, padding=padding) * 2
    elif type == 'line-graymatter':
        img = phantom.line_phantom(fov, padding=padding) * 3
    elif type == 'line-muscle':
        img = phantom.line_phantom(fov, padding=padding) * 4
    elif type == 'line-fat':
        img = phantom.line_phantom(fov, padding=padding) * 6
    else:
        raise ValueError('Incorrect phantom type')

    mask = (img != 0) * 1

    t1 = list(map(lambda x: get_t1(x), img.flatten()))
    t1 = np.array(t1).reshape(img.shape)

    t2 = list(map(lambda x: get_t2(x), img.flatten()))
    t2 = np.array(t2).reshape(img.shape)

    f0 = list(map(lambda x: get_f0(x), img.flatten()))
    f0 = np.array(f0).reshape(img.shape)

    tissue_phantom = {'mask': mask, 't1': t1, 't2': t2, 'f0': f0 }
    return tissue_phantom



import gdown
import os
import zipfile

import mapvbvd
import numpy as np

async def async_read_rawdata(filepath: str, datatype: str='image', doChaAverage: bool=True, doAveAverage: bool=True):
    return read_rawdata(filepath, datatype, doChaAverage, doAveAverage)

def read_rawdata(filepath: str, datatype: str='image', doChaAverage: bool=False, doChaSOSAverage: bool=False, doAveAverage: bool=True, use3dSlices: bool=False):
    """ Reads rawdata files and returns NodeDataset. """
    
    if doChaSOSAverage:
        doChaAverage = False

    twixObj = mapvbvd.mapVBVD(filepath)
    sqzDims = twixObj.image.sqzDims    
    twixObj.image.squeeze = True

    data = twixObj.image['']
    # Move Lin be first index
    linIndex = sqzDims.index('Lin')
    data = np.moveaxis(data, linIndex, 0)
    sqzDims.insert(0, sqzDims.pop(linIndex))

    # Handle averages dimension 
    if doAveAverage and 'Ave' in sqzDims:
        chaIndex = sqzDims.index('Ave')
        data = np.mean(data, axis=chaIndex)
        sqzDims.pop(chaIndex)
                
    # Handle 3d data
    if 'Par' in sqzDims:
        sliceIndex = sqzDims.index('Par')
        data = np.moveaxis(data, sliceIndex, 0)
        sqzDims.insert(0, sqzDims.pop(sliceIndex))
        is3D = True
    else:
        is3D = False

    # Handle fft if required
    if datatype == 'image':
        if is3D:
            data = np.fft.fftshift(np.fft.ifftn(np.fft.fftshift(data, axes=(0, 1, 2))))
        else:
            data = np.fft.fftshift(np.fft.ifft2(np.fft.fftshift(data, axes=(0, 1)), axes=(0, 1)), axes=(0, 1))
    else:  # datatype is kspace
        pass

    # Handle coils dimension - Options for averaging coil data
    if (doChaAverage or doChaSOSAverage) and 'Cha' in sqzDims:
        chaIndex = sqzDims.index('Cha')

        if doChaAverage:
            data = np.mean(data, axis=chaIndex)
        elif doChaSOSAverage:
            data = np.sqrt(np.sum(data**2, axis=chaIndex))
 
        sqzDims.pop(chaIndex)

    # Handle 3D data 
    if 'Sli' in sqzDims:
         # Handle depth - Make data[depth, height, width, ...] Sli = depth
        sliceIndex = sqzDims.index('Sli')
        data = np.moveaxis(data, sliceIndex, 0)
        sqzDims.insert(0, sqzDims.pop(sliceIndex))
    elif use3dSlices: 
        # Make data[depth, height, width, ...] Sli = depth
        # Set depth to 1 if 2d image
        sqzDims.insert(0, 'Sli')
        data = data[np.newaxis, ...]

    mmin = float(np.nanmin(np.abs(data)))
    mmax = float(np.nanmax(np.abs(data)))
    isComplex = np.iscomplexobj(data)

    header = twixObj.hdr
    return {'data': data, 'dims': sqzDims, 'shape': data.shape, 'min': mmin, 'max': mmax, 'isComplex': isComplex}


def download_data(url: str, dataname: str, path: str= os.getcwd()):
    
    # Checks if data folder exists
    targetdir = os.path.join(path, 'data')  
    if not os.path.exists(targetdir):
        os.mkdir(targetdir)

    # Checks if data exists
    filepath = os.path.join(path, 'data', dataname)
    fileExists = os.path.exists(filepath)
    if fileExists:
        print(f'Data: {dataname} data exists')
        return

    # Downloads data
    print('Downloading files ...')
    output = filepath + '.zip'
    gdown.download(url, quiet=False, output=output, fuzzy=True)
    print('Download complete.')

    # Extracts data from .zip 
    print('Extracting files ...')
    with zipfile.ZipFile(output,"r") as zip_ref:
        zip_ref.extractall(targetdir)
        print('Extract complete.')
        print(f'Data located at: ${filepath}')

    # Remove downloaded zip file 
    os.remove(output)



'''Forcing functions to use with GASP.'''

import math
import numpy as np
from scipy import signal
from scipy.stats import norm

def triangle(x0, bw):
    '''Spatial forcing function.

    Parameters
    ----------
    x0 : array_like
        Location (in px).
    bw : float
        Bandwidth of forcing function in Hz, e.g., 1/TR.

    Returns
    -------
    g(x) : complex
        Desired spatial response of uniform phantom.
    '''
    # Naive triangle function implementation
    out = np.zeros(x0.shape)
    for jj, xx in np.ndenumerate(x0):
        if xx < -bw:
            out[jj] = 0
        elif xx > bw:
            out[jj] = 0
        else:
            out[jj] = 1 - np.abs(xx)
    out[np.abs(out) > 0] -= np.min(out)
    return out/np.max(np.abs(out))

def triangle_periodic(img_width, period, offset, bw):
    '''Spatial forcing function - Periodic Triangle with bandwidth

    Parameters
    ----------
    img_width : array_like
        Width of image singal
    period: int
        Period in pixels of forcing function
    offset: int
        Offset in pixels (i.e. cyclic shift)
    bw : float
        Bandwidth (width in pixel from base of triangle) of forcing
        function

    Returns
    -------
    g(x) : complex
        Desired spatial response of uniform phantom with periodic off
        resonance
    '''
    
    bw = round(bw)
    period = round(period)
    assert period >= bw

    window = signal.windows.triang(bw) # signal.triang(bw) <-- scipy v.0.12.0 
    window = np.concatenate((window, np.zeros(period - bw)))
    num_of_windows = math.ceil(img_width / period)
    response = np.tile(window, (num_of_windows))
    response = np.roll(response, offset)
    response = response[:img_width]
    return response

def square(width: int, bw: float, shift: float):
    ''' Square response curve
    width: number of pixels of response
    bw: width of bandpass as fraction of width: 0 to 1
    shift: shift of bandpass as fraction of width: -0.5 to 0.5
    '''
    if bw < 0:
        bw = 0
    if bw > 1:
        bw = 1

    bandpass = np.zeros(width)
    bw = width * bw
    xlo = round(width/2 - bw/2)
    xhi = round(width/2 + bw/2)
    bandpass[xlo:xhi] = 1

    # Shift response
    bandpass = np.roll(bandpass, int(shift * width))
    return bandpass

def sinc(width, bw, shift):
    ''' Sinc response curve
    width: number of pixels of response
    bw: width of bandpass as fraction of width: 0 to 1
    shift: shift of bandpass as fraction of width: -0.5 to 0.5
    '''
    x = np.linspace(-1/bw, 1/bw, width)
    filter = np.sinc(x)

    # Shift response
    filter = np.roll(filter, int(shift * width))
    return filter

def gaussian(width, bw, shift):
    ''' Gaussian response curve 
    width: number of pixels of response
    bw: width of gaussian: 0 to 1
    shift: shift of bandpass as fraction of width: -0.5 to 0.5
    '''
    sigma = bw / .2
    mu = 0
    x = np.linspace(-10, 10, width)
    y = norm.pdf(x, mu, sigma)
    y = np.roll(y, int(shift * width))
    y = y / max(y)
    return y

def notch(width, bw, shift):
    return 1 - gaussian(width, bw, shift)

def make_periodic(x, period:int = 2):
    length = x.shape[0]
    x = np.tile(x,(period))
    #x = np.roll(x, int(length / 2))
    return x

def bandpass_filter(width, bw, shift):
    ''' Bass pass filter
    width: number of pixels of response
    bw: width of bandpass as fraction of width: 0 to 1
    shift: shift of bandpass as fraction of width: -0.5 to 0.5
    '''

    # Parameters
    fs = width * 2  # Sampling frequency (Hz)
    t = np.linspace(0, 1, fs, endpoint=False)  # 1 second of data
    lowcut =  width * (1/2 - bw) # Lower cutoff frequency (Hz)
    highcut = width * (1/2 + bw) # Upper cutoff frequency (Hz)
    lowcut = max(1, lowcut)
    highcut = min(width-1, highcut)
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    order=5

    # Design filter
    b, a = signal.butter(order, [low, high], btype='band')
    #b, a = signal.cheby1(order, 1, [low, high], btype='band')
    #b, a = signal.ellip(order, 1, 40, [low, high], btype='band')

    # Create response
    w, h = signal.freqz(b, a, worN=width)
    h = np.abs(h)

    # Shift response
    h = np.roll(h, int(shift * width))
    return h

def stopband(width, bw, shift):
    ''' Stop band filter
    width: number of pixels of response
    bw: width of bandpass as fraction of width: 0 to 1
    shift: shift of bandpass as fraction of width: -0.5 to 0.5
    '''

    return 1 - bandpass_filter(width, bw, shift)


import os
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation
import seaborn as sns

def view3D(filename: str, data3D: np.ndarray, path: str='./images/') -> None:

    # set seaborn darkgrid theme
    sns.set_theme(style="darkgrid")

    fig, ax = plt.subplots()

    def animate(frame_num):
        ax.clear()
        ax.imshow(abs(data3D[:, :, frame_num]), cmap='gray')
        return ax

    anim = FuncAnimation(fig, animate, frames=data3D.shape[2], interval=1)

    isExist = os.path.exists(path)
    if not isExist:
        os.makedirs(path)

    path = os.path.join(path, filename)
    print(path)
    anim.save(path)



import numpy as np
import matplotlib.pyplot as plt
from gasp import responses, simulation, gasp as GASP

def dixon_3pt(M):
    ''' 3pt dixon - M data with shape [height, width, coils, TEs]'''
    M0 = M[:,:,:,0]
    M1 = M[:,:,:,1]
    M2 = M[:,:,:,2]

    field_map = np.angle(np.conjugate(M0) * M2)
    f = np.exp(-1j * field_map / 2)
    Mw = M0 + M1
    Mf = M0 - M1
    Mw3 = M0 + M1 * np.exp(-1j * field_map / 2)
    Mf3 = M0 - M1 * np.exp(-1j * field_map / 2)

    # 2-pt dixon
    Mw = np.sqrt(np.sum(np.abs(Mw)**2, axis=2))
    Mf = np.sqrt(np.sum(np.abs(Mf)**2, axis=2))

    # 3-pt dixon
    Mw3 = np.sqrt(np.sum(np.abs(Mw3)**2, axis=2))
    Mf3 = np.sqrt(np.sum(np.abs(Mf3)**2, axis=2))

    return Mw3, Mf3

def plot_dixon(Mw, Mf):
    f = plt.figure(figsize=(8,8))
    ax = f.add_subplot(1, 2, 1)
    ax2 = f.add_subplot(1, 2, 2)
    ax.axis('off')
    ax2.axis('off')
    ax.imshow(np.abs(Mw), cmap='gray')
    ax2.imshow(np.abs(Mf), cmap='gray')
    plt.plot()

def plot_dixon_inputdata(M):
    print(M.shape)
    absM = np.sqrt(np.sum(np.abs(M)**2, axis=2))
    f = plt.figure(figsize=(8, 6))
    f.subplots_adjust(wspace=0, hspace=0)

    ax = [f.add_subplot(2,3,i+1) for i in range(6)]
    for a in ax:
        a.axis('off')
        a.set_aspect('equal')

    ax[0].imshow(np.abs(absM[:,:,0]), cmap='gray')
    ax[1].imshow(np.abs(absM[:,:,1]), cmap='gray')
    ax[2].imshow(np.abs(absM[:,:,2]), cmap='gray')
    ax[3].imshow(np.angle(M[:,:,0,0]), cmap='gray')
    ax[4].imshow(np.angle(M[:,:,0,1]), cmap='gray')
    ax[5].imshow(np.angle(M[:,:,0,2]), cmap='gray')
    plt.plot()

def gasp_train(alpha, bw, shift, method="linear"):
    width = 256
    height = 1
    npcs = 16
    TRs = [5e-3, 10e-3, 20e-3]
    alpha = np.deg2rad(alpha)
    gradient = 2 * np.pi
    phantom_type = 'line'
    
    D = responses.gaussian(width, bw=bw, shift=shift)
    M = simulation.simulate_ssfp(width=width, height=height, npcs=npcs, TRs=TRs, alpha=alpha, gradient=gradient, phantom_type=phantom_type, phantom_padding=16)
    Ic, An = simulation.train_gasp(M, D, clines=2, method=method)
    An = np.array(An)[0]
    return Ic, An, M

def gasp_plot_train(alpha = 20, bw = 0.2, shift = 0.01, method="linear"):
    Ic, An, Mtrain = gasp_train(alpha=alpha, bw=bw, shift=shift, method=method)
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10,2))

    _ = np.sqrt(np.sum(np.abs(Mtrain)**2, axis=2))
    _ = abs(_[:,:,0,0])
    print(_.shape)
    ax1.plot(_.T)

    _ = np.abs(Ic)
    print(_.shape)
    ax2.plot(_.T)

def gasp_run_model(Mdata, An, method="linear"):
    Ic = []
    for cc in range(Mdata.shape[2]):
        Mc = Mdata[:,:,cc,:,:]
        Mc = np.reshape(Mc,(Mc.shape[0], Mc.shape[1], -1))
        _ = GASP.apply_gasp(Mc, An, method=method)
        Ic.append(_)
    Ic = np.array(Ic)
    Ic = np.sqrt(np.sum(np.abs(Ic)**2, axis=0))
    return Ic
    
def gasp_train_and_run(Mdata, **options):
    defaultOptions = { 'method':'linear', 'alpha':20, 'shift':20, 'bw':0.2 }
    options = { **defaultOptions, **options }
    
    Ic, An, _ = gasp_train(alpha=options['alpha'], bw=options['bw'], shift=options['shift'], method=options['method'])
    output = gasp_run_model(Mdata, An, method=options['method'])

    plot_gasp_sweep([[output, Ic]])
    return output, Ic, An

def gasp_sweep(Mdata, sweep_type = 'shift', sweep_start=-0.5, sweep_end=0.5, sweep_size=10, plot_type = 'gasp', **options):
    ''' Plots a sweep of gasp images using sweep_type. You can sweep alpha, shift and bw. '''
    
    defaultOptions = { 'method':'linear', 'alpha':20, 'shift':20, 'bw':0.2 }
    options = { **defaultOptions, **options }

    dataset = []
    sweep_values = np.linspace(sweep_start, sweep_end, sweep_size)
    for val in sweep_values:
        alpha = val if sweep_type == 'alpha' else options['alpha']
        bw = val if sweep_type == 'bw' else options['bw']
        shift = val if sweep_type == 'shift' else options['shift']

        Ic, An, _ = gasp_train(alpha=alpha, bw=bw, shift=shift, method=options['method'])
        output = gasp_run_model(Mdata, An, method=options['method'])
        dataset.append([output, Ic])

    if (plot_type == 'all'):
        plot_gasp_sweep_with_training(dataset)
    if (plot_type == 'gasp'):
        plot_gasp_sweep(dataset)

def plot_gasp_sweep(dataset):
    length = len(dataset)
    f = plt.figure(figsize=(20,6))
    for i in range(length):
        data0 = dataset[i][0]
        ax = f.add_subplot(1, length, i + 1)
        ax.imshow(data0, cmap='gray')
        ax.axis('off')
    plt.show()

def plot_gasp_sweep_with_training(dataset):
    length = len(dataset)
    f = plt.figure(figsize=(20,6))
    for i in range(length):
        data0 = dataset[i][0]
        data1 = dataset[i][1]
        ax = f.add_subplot(2, length, i + 1)
        ax2 = f.add_subplot(2, length, length + i+1)
        ax.imshow(data0, cmap='gray')
        ax.axis('off')
        ax2.plot(np.abs(data1).T)
        ax2.axis('off')
    plt.show()


"""GASP module."""

import numpy as np
import numpy.typing as npt
from skimage.filters import threshold_li
from scipy.optimize import least_squares

def run_gasp(I: npt.NDArray, An: npt.NDArray, method :str = "affine"):
    ''' Run GASP model on data with shape [Height, Width, PC x TRs] 
    
    Parameters:
    I (NDArray): Array of phase-cycled images.
    D (NDArray): Vector of samples of desired spectral profile.
    method (str, optional): Method used to compute the GASP model solution.
        Must be one of {"affine", "linear", "lev-mar", "lev-mar-quad"}.

    Reuturns:
    NDArray: Reconstructed image.
    '''

    # Reshape the data to be in the form [Height x Width, PCs x TRs]
    height, width = I.shape[:2]
    I = I.reshape(I.shape[0], I.shape[1], -1)   # Collapse all but first 2 dimensions
    I = I.reshape((-1, I.shape[-1]))            # Collapse all dimensions last dimension
    npcs = I.shape[-1]
    data_shape = (height, width)

    if method == "linear":
        out = I.dot(An).reshape(data_shape)
    elif method == "affine":
        I = np.column_stack((np.ones(I.shape[0]), I))
        out = I.dot(An).reshape(data_shape)
    elif method == "quad":
        I = np.column_stack((np.ones(I.shape[0]), I, I**2))
        out = I.dot(An).reshape(data_shape)
    elif method == "levmar-old":
        x0 = An[:npcs]              # Linear terms
        x1 = An[npcs:]              # Quadratic terms
        out = np.reshape(I @ x0 + I**2 @ x1, data_shape)
    elif method == "levmar-quad":
        c = An[0]                    # Constant term
        x0 = An[1:npcs+1]            # Linear terms
        x1 = An[npcs+1:]             # Quadratic terms
        out = np.reshape(c + I @ x0 + I**2 @ x1, data_shape)
    else:
        raise ValueError(f"method '{method}' was not recognized")

    return out

def train_gasp(I: npt.NDArray, D: npt.NDArray, method: str = "affine"):
    ''' Train GASP model on data with shape [Height, Width, PCs x TRs] and desired spectral profile D with shape [Width,]
    
    Parameters:
    I (NDArray): Array of phase-cycled images.
    D (NDArray): Vector of samples of desired spectral profile.
    method (str, optional): Method used to compute the GASP model solution.
        Must be one of {"affine", "linear", "lev-mar", "lev-mar-quad"}.

    Reuturns:
    tuple: Reconstructed image and coefficients.
    '''
 
    # Reshape the data to be in the form [Height x Width, PCs x TRs]
    height, width = I.shape[:2]
    I = I.reshape(I.shape[0], I.shape[1], -1)   # Collapse all but first 2 dimensions
    I = I.reshape((-1, I.shape[-1]))            # Collapse all dimensions last dimension
    npcs = I.shape[-1]
    data_shape = (height, width)

    # Repeat the desired spectral profile to match the number of PCs
    D = np.tile(D, (int(I.shape[0]/D.size),))

    # Now solve the system
    if method == "linear":
        A = np.linalg.lstsq(I, D, rcond=None)[0]  # Solves a linear system of form: D = A * I (i.e. y = a * x)
        out = I.dot(A).reshape(data_shape)     # Reconstruct the image from the coefficients 
    elif method == "affine":
        I = np.column_stack((np.ones(I.shape[0]), I))  # Add a column of ones (b) to the data so data is from of y = a * x + b
        A = np.linalg.lstsq(I, D, rcond=None)[0]  # Solves a linear system of form: D = A * I (i.e. y = a * x)
        out = I.dot(A).reshape(data_shape)     # Reconstruct the image from the coefficients 
    elif method == "quad":
        I_quad = np.column_stack((np.ones(I.shape[0]), I, I**2))  # Add columns for constant, linear, and quadratic terms
        A = np.linalg.lstsq(I_quad, D, rcond=None)[0]  # Solves a quadratic system of form: D = a * I^2 + b * I + c
        out = I_quad.dot(A).reshape(data_shape) 
    elif method == 'levmar-old':
        def residuals(y):
            y0 = y[:npcs] + 1j*y[npcs:2*npcs]
            y1 = y[2*npcs:3*npcs] + 1j*y[3*npcs:4*npcs]
            residual = I @ y0 + I**2 @ y1 - D
            return np.concatenate((residual.real, residual.imag))

        res = least_squares(fun=residuals, x0=np.zeros(npcs*4), method="lm")
        if not res.success:
            print(f"GASP SOLVE ERROR ({method}): {res.message}")

        x0 = res.x[:npcs] + 1j*res.x[npcs:2*npcs]
        x1 = res.x[2*npcs:3*npcs] + 1j*res.x[3*npcs:4*npcs]
        A = np.concatenate((x0, x1))
        out = np.reshape(I @ x0 + I**2 @ x1, data_shape)
    elif method == 'levmar-quad':
        def quadratic_model_residuals(params):
            c = params[0] + 1j*params[1]                                    # Constant term
            y0 = params[2:npcs+2] + 1j*params[npcs+2:2*npcs+2]              # Linear terms
            y1 = params[2*npcs+2:3*npcs+2] + 1j*params[3*npcs+2:4*npcs+2]   # Quadratic terms
            residual = c + I @ y0 + I**2 @ y1 - D
            return np.concatenate((residual.real, residual.imag))

        # Initialize parameters
        initial_params = np.zeros(4*npcs + 2)
        initial_params[0] = np.real(np.mean(D))  # Real part of constant
        initial_params[1] = np.imag(np.mean(D))  # Imaginary part of constant

        res = least_squares(fun=quadratic_model_residuals, x0=initial_params, method="lm")
        if not res.success:
            print(f"GASP SOLVE ERROR ({method}): {res.message}")
        
        c = res.x[0] + 1j*res.x[1]
        x0 = res.x[2:npcs+2] + 1j*res.x[npcs+2:2*npcs+2]
        x1 = res.x[2*npcs+2:3*npcs+2] + 1j*res.x[3*npcs+2:4*npcs+2]
        A = np.concatenate(([c], x0, x1))
        out = np.reshape(c + I @ x0 + I**2 @ x1, data_shape)
    else:
        raise ValueError(f"method '{method}' was not recognized")
    
    A = np.array(A)
    return out, A

def train_gasp_with_coils(data, D, method="affine"):

    # Get dimensions
    if data.ndim == 4:
        height, width, ncoils, npcs = data.shape[:]
    if data.ndim == 5:
        height, width, ncoils, npcs, TRs = data.shape[:]
        npcs = npcs * TRs       # Combine PCs and TRs

    # Run gasp
    out = np.zeros((ncoils, height, width), dtype='complex')
    n = npcs+1 if method == 'affine' else npcs
    An = np.zeros((ncoils, n), dtype='complex')
    for cc in range(ncoils):
        single_coil = data[:,:, cc, ...].reshape(height, width, npcs)
        out[cc, ...], An[cc, ...] = train_gasp(single_coil, D, method=method)
    out = np.sqrt(np.sum(np.abs(out)**2, axis=0))
    
    return out, An

def create_data_mask(M):
    # Create mask of phantom
    _ = np.sqrt(np.sum(np.abs(M)**2, axis=2))
    _ = np.mean(_, axis=2)
    _ = np.mean(_, axis=2)
    _ = abs(_)
    thresh = threshold_li(_)
    mask = np.abs(_) > thresh
    return mask

def apply_mask_to_data(M, mask):
    # Ensure mask has the same shape as the first two dimensions of M
    if mask.shape != M.shape[:2]:
        raise ValueError("Mask shape does not match the first two dimensions of the data")

    if len(M.shape) == 5: 
        # Expand mask to match the dimensions of M using np.tile
        expanded_mask = mask[:, :, np.newaxis, np.newaxis, np.newaxis]
        expanded_mask = np.tile(expanded_mask, (1, 1, M.shape[2], M.shape[3], M.shape[4]))

    # Apply the mask
    masked_data = M * expanded_mask

    return masked_data

def extract_centered_subset(data, n_lines):
    """
    Extract a subset of the data centered around the middle of the height dimension.
    
    Args:
    data (np.ndarray): Input data with shape [height, width, coils, pcs, TR]
    n_lines (int): Number of lines to extract in the height dimension
    
    Returns:
    np.ndarray: Subset of the data with shape [n_lines, width, coils, pcs, TR]
    """
    if n_lines > data.shape[0]:
        raise ValueError("n_lines cannot be greater than the height of the data")
    
    # Calculate the start and end indices for the subset
    center = data.shape[0] // 2
    start = center - n_lines // 2
    end = start + n_lines
    
    # Extract the subset
    subset = data[start:end, ...]
    
    return subset

def process_data_for_gasp(M, useMask=False, useCalibration=False, n_lines=2):
    ''' Process data for GASP model training.
    
    Parameters:
    M (NDArray): Array of phase-cycled images.
    D (NDArray): Vector of samples of desired spectral profile.
    useMask (bool, optional): Whether to use a mask to remove background from the data.
    useCalibration (bool, optional): Whether to use a calibration region.
    n_lines (int, optional): Number of lines to extract from the center of the data.
    
    Returns:
    NDArray: Processed data.
    '''

    if useMask:
        mask = create_data_mask(M)
        data = apply_mask_to_data(M, mask)
    else:
        data = M

    if useCalibration:
        data = extract_centered_subset(data, n_lines)

    return data



import numpy as np

def phantom_generator(fov: int=256, coil: int=1, type: str='shepp_logan', padding: int = 8) -> np.ndarray:
    """ Generates a phantom with a given shape for a number of coils.
    Args:
        fov: size of image.
        coil: number of coils (Dtype): data type.
        type: type of phantom
    Returns:
        array.
    """
    fov = int(fov)
    coil = int(coil)

    if type == 'shepp_logan':
        rawdata = shepp_logan_phantom([fov, fov])
    elif type == 'circle':
        rawdata = circle_phantom([fov, fov])
    elif type == 'circles':
        rawdata = circle_array_phantom([fov, fov])
    elif type == 'block':
        rawdata = block_phantom_single(fov, padding=padding)
    elif type == 'blocks':
        rawdata = block_phantom(fov, padding=padding)
    elif type == 'line':
        rawdata = line_phantom(fov, padding=padding)
    else:
        raise ValueError('Incorrect phantom type')

    img = np.repeat(rawdata[:, :, np.newaxis], coil, axis=2)
    #img = img * generate_birdcage_sensitivities(matrix_size=fov, number_of_coils=coil)
    return img

def line_phantom(length: int=256, padding: int=32):
    s = (1, length) 
    line = np.ones(s) * 1.0

    if padding > 0:
        line[0, -1] = 0
        line[0, -padding:-1] = 0
        line[0, 0:padding] = 0
        
    line.astype(int)
    return line 

def block_phantom_single(shape: int=256, padding: int=32):
    s = (shape - 2 * padding, shape - 2 * padding)
    block = np.ones(s) * 1.0
    block = np.pad(block, (padding, padding))
    block.astype(int)
    return block

def block_phantom(shape: int=256, blocks: int=8, padding: int=8):
    width = 64
    height = 64

    s = (width - 2 * padding, height - 2 * padding)
    patches = [[], []]
    for i in range(blocks+1):
        if i > 0:
            patch = np.ones(s) * i
            patch = np.pad(patch, (padding, padding))
            patches[int((i - 1) / 4)].append(patch)
        
    mask = np.block(patches)
    mask = mask.astype(int)
    mask = np.pad(mask, ((64, 64), (0, 0)))

    return mask


def generate_birdcage_sensitivities(matrix_size: int=256, number_of_coils: int=8, relative_radius: float=1.5, normalize: bool=True) -> np.ndarray:
    """ Generates birdcage coil sensitivities.
    :param matrix_size: size of imaging matrix in pixels (default ``256``)
    :param number_of_coils: Number of simulated coils (default ``8``)
    :param relative_radius: Relative radius of birdcage (default ``1.5``)
    This function is heavily inspired by the mri_birdcage.m Matlab script in
    Jeff Fessler's IRT package: http://web.eecs.umich.edu/~fessler/code/
    """

    out = np.zeros((number_of_coils, matrix_size, matrix_size), dtype=np.complex64)
    for c in range(0, number_of_coils):
        coilx = relative_radius*np.cos(c*(2*np.pi/number_of_coils))
        coily = relative_radius*np.sin(c*(2*np.pi/number_of_coils))
        coil_phase = -c*(2*np.pi/number_of_coils)

        for y in range(0, matrix_size):
            y_co = float(y-matrix_size/2)/float(matrix_size/2)-coily
            for x in range(0, matrix_size):
                x_co = float(x-matrix_size/2)/float(matrix_size/2)-coilx
                rr = np.sqrt(x_co**2+y_co**2)
                phi = np.arctan2(x_co, -y_co) + coil_phase
                out[c, y, x] = (1/rr) * np.exp(1j*phi)
    if normalize:
         rss = np.squeeze(np.sqrt(np.sum(abs(out) ** 2, 0)))
         out = out / np.tile(rss, (number_of_coils, 1, 1))
    out = np.moveaxis(out, 0, -1)
    return out


def shepp_logan_phantom(fov):
    """ Generates a Shepp Logan phantom with a given shape.
    Args:
        fov: size
    Returns:
        array.
    """

    sl_amps = [1, -0.8, -0.2, -0.2, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    sl_scales = [[.6900, .920, .810],  # white big
                 [.6624, .874, .780],  # gray big
                 [.1100, .310, .220],  # right black
                 [.1600, .410, .280],  # left black
                 [.2100, .250, .410],  # gray center blob
                 [.0460, .046, .050],
                 [.0460, .046, .050],
                 [.0460, .046, .050],  # left small dot
                 [.0230, .023, .020],  # mid small dot
                 [.0230, .023, .020]]

    sl_offsets = [[0., 0., 0],
                 [0., -.0184, 0],
                 [.22, 0., 0],
                 [-.22, 0., 0],
                 [0., .35, -.15],
                 [0., .1, .25],
                 [0., -.1, .25],
                 [-.08, -.605, 0],
                 [0., -.606, 0],
                 [.06, -.605, 0]]

    sl_angles = [[0, 0, 0],
                 [0, 0, 0],
                 [-18, 0, 10],
                 [18, 0, 10],
                 [0, 0, 0],
                 [0, 0, 0],
                 [0, 0, 0],
                 [0, 0, 0],
                 [0, 0, 0],
                 [0, 0, 0]]

    image = phantom(fov, sl_amps, sl_scales, sl_offsets, sl_angles, dtype=complex)
    return image


def circle_array_phantom(fov):
    """ Generates an array of circle phantoms with a given shape and dtype.
    Args:
        fov: size
    Returns:
        array.
    """
    amps = [1, 2, 3, 4]
    scales = [[0.25, 0.25, 0.25], [0.25, 0.25, 0.25], [0.25, 0.25, 0.25], [0.25, 0.25, 0.25]]
    offsets = [[0.5, 0.5, 0], [-0.5, 0.5, 0], [0.5, -0.5, 0], [-0.5, -0.5, 0]]
    angles = [[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]]

    image = phantom(fov, amps, scales, offsets, angles, dtype=float)
    return image


def circle_phantom(fov):
    """ Generates a circle phantom with a given size.
    Args:
        fov: size
    Returns:
        array.
    """
    amps = [1]
    scales = [[0.75, 0.75, 0.75]]
    offsets = [[0, 0, 0]]
    angles = [[0, 0, 0]]

    image = phantom(fov, amps, scales, offsets, angles, dtype=float)
    return image


def phantom(shape, amps, scales, offsets, angles, dtype):
    """
    Generate a cube of given shape using a list of ellipsoid
    parameters.
    """
    if len(shape) == 2:
        ndim = 2
        shape = (1, shape[-2], shape[-1])

    elif len(shape) == 3:
        ndim = 3

    else:
        raise ValueError('Incorrect dimension')

    out = np.zeros(shape, dtype=dtype)

    z, y, x = np.mgrid[-(shape[-3] // 2):((shape[-3] + 1) // 2),
                       -(shape[-2] // 2):((shape[-2] + 1) // 2),
                       -(shape[-1] // 2):((shape[-1] + 1) // 2)]

    coords = np.stack((x.ravel() / shape[-1] * 2,
                       y.ravel() / shape[-2] * 2,
                       z.ravel() / shape[-3] * 2))

    for amp, scale, offset, angle in zip(amps, scales, offsets, angles):
        ellipsoid(amp, scale, offset, angle, coords, out)

    if ndim == 2:
        return out[0, :, :]
    else:
        return out


def ellipsoid(amp, scale, offset, angle, coords, out):
    """
    Generate a cube containing an ellipsoid defined by its parameters.
    If out is given, fills the given cube instead of creating a new
    one.
    """
    R = rotation_matrix(angle)
    coords = (np.matmul(R, coords) - np.reshape(offset, (3, 1))) / \
        np.reshape(scale, (3, 1))

    r2 = np.sum(coords ** 2, axis=0).reshape(out.shape)

    out[r2 <= 1] += amp


def rotation_matrix(angle):
    """ Generate rotation matrix """
    cphi = np.cos(np.radians(angle[0]))
    sphi = np.sin(np.radians(angle[0]))
    ctheta = np.cos(np.radians(angle[1]))
    stheta = np.sin(np.radians(angle[1]))
    cpsi = np.cos(np.radians(angle[2]))
    spsi = np.sin(np.radians(angle[2]))
    alpha = [[cpsi * cphi - ctheta * sphi * spsi,
              cpsi * sphi + ctheta * cphi * spsi,
              spsi * stheta],
             [-spsi * cphi - ctheta * sphi * cpsi,
              -spsi * sphi + ctheta * cphi * cpsi,
              cpsi * stheta],
             [stheta * sphi,
              -stheta * cphi,
              ctheta]]
    return np.array(alpha)



from .__about__ import __version__
from .path import get_project_path, get_gasp_path
from .gasp import run_gasp, train_gasp, train_gasp_with_coils, process_data_for_gasp, create_data_mask, apply_mask_to_data, extract_centered_subset
from .simulation import SSFPParams, simulate_ssfp, simulate_ssfp_sampling, simulate_ssfp_simple



""" ssfp simulation """

import numpy as np


def ssfp(T1, T2, TR, TE, alpha, dphi=(0,), field_map=0, M0=1, f0=0, phi=0, useSqueeze=True) -> np.ndarray:
    """ Multiple acquisition ssfp """
    dphi = np.atleast_2d(dphi)

    M = []
    for ii, pc in np.ndenumerate(dphi):
        M.append(_ssfp(T1, T2, TR, TE, alpha, pc, field_map, M0, f0, phi)[..., None])
    M = np.concatenate(M, axis=-1)
    
    # Squeeze out dim of length 1, otherwise shape is [width, height, dphi]
    M = np.squeeze(M)
    if not useSqueeze and len(dphi) == 1:
        M = M[..., None]
    return M


def _ssfp(T1, T2, TR, TE, alpha, dphi=0, field_map=0, M0=1, f0=0, phi=0):
    """ transverse signal for ssfp mri after excitation at TE

    Parameters
    ----------
    T1 : float or array_like
        longitudinal exponential decay time constant (in seconds).
    T2 : float or array_like
        transverse exponential decay time constant (in seconds).
    TR : float
        repetition time (in seconds).
    alpha : float or array_like
        flip angle (in rad).
    dphi : float, optional
        Linear phase-cycle increment (in rad).
    field_map : float or array_like, optional
        B0 field map (in Hz).
    M0 : float or array_like, optional
        proton density.
    f0 : float, optional
        off-resonance (in Hz). Includes factors like the chemical shift
        of species w.r.t. the water peak.
    phi : float, optional
        phase offset (in rad).
    """
    
    # Convention for Ernst-Anderson based implementation from Hoff
    field_map = -1 * field_map
    
    # Set T1, T2, alpha, and field_map inputs to arrays
    T1 = np.atleast_2d(T1)
    T2 = np.atleast_2d(T2)
    alpha = np.atleast_2d(alpha)
    f0 = np.atleast_2d(f0)
    field_map = np.atleast_2d(field_map)

    # Compute exponential decay and handle T1, T2 of zero
    E1 = np.zeros(T1.shape)
    E1[T1 > 0] = np.exp(-TR/T1[T1 > 0])
    E2 = np.zeros(T2.shape)
    E2[T2 > 0] = np.exp(-TR/T2[T2 > 0])

    # Precompute theta and derivatives of theta and alpha
    beta = 2 * np.pi * (f0 + field_map) * TR
    theta = beta - dphi; # theta => phase per repetition time
    cos_alpha = np.cos(alpha)
    sin_alpha = np.sin(alpha)
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    # Calculate Mxy 
    Mbottom = (1 - E1 * cos_alpha) * (1 - E2 * cos_theta) - E2 * (E1 - cos_alpha) * (E2 - cos_theta)
    Mx = M0 * (1 - E1) * sin_alpha * (1 - E2 * cos_theta) / Mbottom
    My = M0 * (1 - E1) * E2 * sin_alpha * sin_theta / Mbottom
    Mc = Mx + 1j * My

    # Add additional phase and handle T2 of zero 
    T2 = np.array(T2)
    idx = np.where(T2 > 0)
    val = np.zeros(T2.shape)
    val[idx] = -TE/T2[idx]
    _phi = beta * (TE / TR) + phi
    Mc = Mc * np.exp(1j * _phi) * np.exp(val)

    return Mc


def add_noise(I: np.ndarray, mu: float=0, sigma: float=0.005, factor: float=1) -> np.ndarray:
    """add gaussian noise to given simulated bSSFP signals

    Parameters
    ----------
    I: array_like
       images size(M,N,C)
    mu: float
        mean of the normal distribution
    sd: float
        standard deviation of the normal distribution

        Returns
    -------
    Mxy : numpy.ndarray
        Transverse complex magnetization with added .
    """
    noise = factor * np.random.normal(mu, sigma, (2,) + np.shape(I))
    noise_matrix = noise[0] + 1j*noise[1]
    return I + noise_matrix



import math
import numpy as np
import matplotlib.pyplot as plt 
from skimage.filters import threshold_li
from gasp import ssfp, tissue, gasp as GASP
import warnings
warnings.simplefilter('ignore')


def simulate_ssfp(width = 256, height = 256, npcs = 16, TRs = [5e-3, 10e-3, 20e-3], alpha = np.deg2rad(60), gradient = 2 * np.pi, gradientOffset = 0, f0 = 0, phantom_type='circle', minTR=None, useSqueeze: bool=True, pcs=None, phantom_padding=8):
    ''' Simulates bssfp with tissue phantom '''

    # Create phantoms, tissues, parameters
    t = tissue.tissue_generator(type=phantom_type, padding=phantom_padding)
    mask = t['mask']
    size = mask.shape
    t1 = t['t1']
    t2 = t['t2']
    BetaMax = gradient
    beta = np.linspace(-BetaMax, BetaMax, size[1]) + gradientOffset
    if minTR is None:
        minTR = TRs[0]
    f = beta / minTR / (2 * np.pi) 
    f = np.tile(f, (size[0], 1))
    f = f + t['f0'] + f0

    # use explicitly provided PCs if given, otherwise assume linear distribution with npcs points
    if pcs is None:
        assert isinstance(npcs, int)
        pcs = np.linspace(0, 2*np.pi, npcs, endpoint=False)

    # Create simulated phantom data
    nTRs = len(TRs)
    M = np.empty((height, width, npcs,  nTRs), dtype=np.complex128)
    for ii, TR in enumerate(TRs):
        TE = TR / 2.0
        M[..., ii] = ssfp.ssfp(t1, t2, TR, TE, alpha, pcs, field_map=f, M0 = mask, useSqueeze=useSqueeze)
    M = np.reshape(M, (height, width, 1, npcs,  nTRs))
    #M = ssfp.add_noise(M, sigma=0.005)
    return M

class SSFPParams():
    def __init__(self, length, alpha, TRs, pcs):
        
        self.length = length

        if isinstance(alpha, float):
            self.alpha = np.ones(length) * alpha
        else:
            self.alpha = alpha
        self.alpha = np.array(self.alpha).reshape(-1, 1) 
        
        self.TRs = np.array(TRs).reshape(-1, 1)
        self.pcs = np.array(pcs).reshape(-1, 1)

    def get(self, index):
        return self.alpha[index], self.TRs[index], self.pcs[index]

    def getAlpha(self, index):
        return self.alpha[index]

    def getTR(self, index):
        return self.TRs[index]

    def getPC(self, index):
        return self.pcs[index]

def simulate_ssfp_sampling(width = 256, height = 256, params=None, minTR = 5e-3, gradient = 2 * np.pi, phantom_type='circle', useSqueeze: bool=True, phantom_padding=8):
    ''' Simulates bssfp with tissue phantom. Uses a list of parameters to generate phantoms with a set of alpha, pcs, TRs parameters. '''
    
    # Create phantoms, tissues, parameters
    t = tissue.tissue_generator(fov=width, type=phantom_type, padding=phantom_padding)
    mask = t['mask']
    size = mask.shape
    t1 = t['t1']
    t2 = t['t2']
    f0 = t['f0']
    BetaMax = gradient
    beta = np.linspace(-BetaMax, BetaMax, size[1])
    f = beta / minTR / (2 * np.pi) 
    f = np.tile(f, (size[0], 1))
    f = f + f0

    # Create simulated phantom data
    npcs = params.length
    M = np.empty((height, width, npcs), dtype=np.complex128)
    for ii in range(npcs):
        alpha = params.getAlpha(ii)
        pcs = params.getPC(ii)
        TR = params.getTR(ii)
        TE = TR / 2.0
        M[..., ii] = ssfp.ssfp(t1, t2, TR, TE, alpha, pcs, field_map=f, M0 = mask, useSqueeze=useSqueeze)
    M = np.reshape(M, (height, width, npcs))
    #M = ssfp.add_noise(M, sigma=0.005)
    return M

def simulate_ssfp_simple(width = 256, height = 1,  T1=.035, T2 =.035, params=None, minTR = 5e-3, gradient = 2 * np.pi, f0 = 0):
    ''' Simulates bssfp with tissue phantom. Uses a list of parameters to generate data with a set of alpha, pcs, TRs parameters. '''
    
    # Parameters
    size = [height, width]
    BetaMax = gradient
    beta = np.linspace(-BetaMax, BetaMax, size[1])
    f = beta / minTR / (2 * np.pi) 
    f = np.tile(f, (size[0], 1))
    f = f + f0

    # Create simulated phantom data
    npcs = params.length
    M = np.empty((height, width, npcs), dtype=np.complex128)
    for ii in range(npcs):
        alpha = params.getAlpha(ii)
        pcs = params.getPC(ii)
        TR = params.getTR(ii)
        TE = TR / 2.0
        M[..., ii] = ssfp.ssfp(T1, T2, TR, TE, alpha, pcs, field_map=f)
    M = np.reshape(M, (height, width, npcs))
    #M = ssfp.add_noise(M, sigma=0.005)
    return M

def train_gasp(M, D, clines=32, method:str = 'linear'):
    ''' Deprecated '''

    # Create mask of phantom
    _ = np.sqrt(np.sum(np.abs(M)**2, axis=2))
    _ = np.mean(_, axis=2)
    _ = np.mean(_, axis=2)
    _ = abs(_)
    thresh = threshold_li(_)
    mask = np.abs(_) > thresh

    # Apply mask to data
    mask0 = np.tile(
        mask, (M.shape[2:] + (1, 1,))).transpose((3, 4, 0, 1, 2))
    data = M * mask0
    
    # Reshape data to required shapes
    data = np.reshape(data, data.shape[:-2] + (-1,))    # [Height, Width, Coil, PCs x TRs]
    data = np.moveaxis(data, 2, 0)                      # [Coil, Height, Width, PCs x TRs]
    data = data.transpose((0, 3, 1, 2))                 # [Coil,  PCs x TRs, Width, Height]

    # Get new dimensions
    ncoils, npcs, height, width = data.shape[:]

    # Calibration box - (# Number of lines of calibration, Pixels on signal)
    C_dim = (clines, width)
    
    # Run gasp
    Ic = np.zeros((ncoils, height, width), dtype='complex')
    An = []
    for cc in range(ncoils):
        Ic[cc, ...], _An = GASP.gasp_coefficients(data[cc, ...], D, C_dim, pc_dim=0, method=method)
        An.append(_An)
    Ic = np.sqrt(np.sum(np.abs(Ic)**2, axis=0))

    return Ic, An


def evaluate_gasp(M, G, method:str = "linear"):
    Ic = GASP.apply_gasp(M, G, method=method)
    return Ic


def simulate_gasp(D, width = 256, height = 256, npcs = 16, TRs = [5e-3, 10e-3, 20e-3], alpha = np.deg2rad(60), gradient = 2 * np.pi, phantom_type='circle'):
    ''' Simulates gasp with tissue phantom '''

    # Simulate ssfp with tissue phantom 
    M = simulate_ssfp(width=width, height=height, npcs=npcs, TRs=TRs, alpha=alpha, gradient=gradient, phantom_type=phantom_type)
    
    # Train gasp model coefficients
    Ic, An = train_gasp(M, D)
    
    return Ic, M, An


def view_gasp_input(M, slices=[0,0]):
    ''' Plots input magnetization, M'''
    
    height = M.shape[0]
    width = M.shape[1]

    # Plot data
    _ = np.sqrt(np.sum(np.abs(M)**2, axis=2))
    _ = abs(_[:,:,slices[0],slices[1]])

    f = plt.figure(figsize=(20,3))
    ax = f.add_subplot(121)
    ax2 = f.add_subplot(122)

    ax.imshow(_, cmap='gray')
    ax2.plot(_[int(width/2), :])
    plt.show()


def view_gasp_results(Ic, M, D):
    ''' Plots the results of gasp for given gasp output, Ic, input magnetization, M, 
        and target spectrum, D '''
    
    height = M.shape[0]
    width = M.shape[1]

    # Plot data
    _ = np.sqrt(np.sum(np.abs(M)**2, axis=2))
    _ = abs(_[:,:,0,0])

    f = plt.figure(figsize=(20,3))
    ax = f.add_subplot(141)
    ax2 = f.add_subplot(142)
    ax3 = f.add_subplot(143)
    ax4 = f.add_subplot(144)

    ax.imshow(_, cmap='gray')
    ax2.plot(_[int(width/2), :])
    ax3.imshow(Ic, cmap='gray')
    ax4.plot(np.abs(Ic[int(height/2), :]), label='Simulated Profile')
    ax4.plot(D, '--', label='Desired Profile')

    plt.show()


def view_gasp(Ic, D):
    Ic = np.abs(Ic)
    s = np.abs(Ic[int(Ic.shape[0]/2), :])

    f = plt.figure(figsize=(10,6))
    ax = f.add_subplot(2, 2, 1)
    ax2 = f.add_subplot(2, 2, 2)
    ax.imshow(Ic, cmap='gray')
    ax2.plot(s, label='Simulated Profile')
    ax2.plot(D, '--', label='Desired Profile')


def view_gasp_comparison(G, D):
    G = np.abs(G)
    length = len(G)

    f = plt.figure(figsize=(20,6))
    for i in range(length):
        ax = f.add_subplot(2, 8, i+1)
        ax2 = f.add_subplot(2, 8, 8 + i+1)

        g = G[i]
        d = D[i]
        ax.imshow(g, cmap='gray')
        ax.axis('off')
        ax2.plot(np.abs(g[int(g.shape[0]/2), :]), label='Simulated Profile')
        ax2.plot(d, '--', label='Desired Profile')



import math
import numpy as np
import matplotlib.pyplot as plt

def spiral_sampling(n_points, n_turns, tr_range=(5e-3, 20e-3)):
    theta = np.linspace(0, 2*np.pi*n_turns, n_points)
    r = np.linspace(0, 1, n_points)
    min_TR, max_TR = tr_range
    TR = min_TR + r * (max_TR - min_TR)
    PC = theta % (2*np.pi)
    return TR, PC

def fibonacci_sampling(n_points, tr_range=(5e-3, 20e-3)):
    golden_ratio = (1 + 5**0.5) / 2
    i = np.arange(n_points)
    theta = 2 * np.pi * i / golden_ratio
    r = np.sqrt(i / n_points)
    min_TR, max_TR = tr_range
    TR = min_TR + r * (max_TR - min_TR)
    PC = theta % (2*np.pi)
    return TR, PC

def log_polar_sampling(n_points, tr_range=(5e-3, 20e-3)):
    n_radial = int(np.sqrt(n_points))
    n_angular = n_points // n_radial
    
    min_TR, max_TR = tr_range
    log_min, log_max = np.log(min_TR), np.log(max_TR)
    r = np.exp(np.linspace(log_min, log_max, n_radial))
    theta = np.linspace(0, 2*np.pi, n_angular, endpoint=False)
    R, Theta = np.meshgrid(r, theta)
    TR, PC = R.flatten(), Theta.flatten()
    
    if len(TR) > n_points:
        TR, PC = TR[:n_points], PC[:n_points]
    elif len(TR) < n_points:
        extra = n_points - len(TR)
        TR = np.pad(TR, (0, extra), mode='edge')
        PC = np.pad(PC, (0, extra), mode='edge')
    
    return TR, PC

def grid_sampling(n_points, n_turns, tr_range=(5e-3, 20e-3)):
    min_TR, max_TR = tr_range
    n_tr = n_turns
    n_pc = n_points // n_tr
    
    TRs = np.linspace(min_TR, max_TR, n_tr)
    PCs = np.linspace(0, 2*np.pi, n_pc, endpoint=False)
    
    TR, PC = np.meshgrid(TRs, PCs)
    TR, PC = TR.flatten(), PC.flatten()
    
    if len(TR) > n_points:
        TR, PC = TR[:n_points], PC[:n_points]
    elif len(TR) < n_points:
        extra = n_points - len(TR)
        TR = np.pad(TR, (0, extra), mode='edge')
        PC = np.pad(PC, (0, extra), mode='edge')
    
    return TR, PC

def grid_multiples_sampling(n_points, n_turns, tr_range=(5e-3, 20e-3)):
    min_TR, max_TR = tr_range
    n_tr = n_turns
    n_pc = n_points // n_tr
    
    TRs = min_TR * np.linspace(1, n_turns, n_tr)
    PCs = np.linspace(0, 2*np.pi, n_pc, endpoint=False)
    
    TR, PC = np.meshgrid(TRs, PCs)
    TR, PC = TR.flatten(), PC.flatten()
    
    if len(TR) > n_points:
        TR, PC = TR[:n_points], PC[:n_points]
    elif len(TR) < n_points:
        extra = n_points - len(TR)
        TR = np.pad(TR, (0, extra), mode='edge')
        PC = np.pad(PC, (0, extra), mode='edge')
    
    return TR, PC

def grid_pow2_sampling(n_points, n_turns, tr_range=(5e-3, 20e-3)):
    min_TR, max_TR = tr_range
    n_tr = n_turns
    n_pc = n_points // n_tr
    
    TRs = min_TR * 2 ** np.linspace(0, n_turns, n_tr, endpoint=False)
    PCs = np.linspace(0, 2*np.pi, n_pc, endpoint=False)
    
    TR, PC = np.meshgrid(TRs, PCs)
    TR, PC = TR.flatten(), PC.flatten()
    
    if len(TR) > n_points:
        TR, PC = TR[:n_points], PC[:n_points]
    elif len(TR) < n_points:
        extra = n_points - len(TR)
        TR = np.pad(TR, (0, extra), mode='edge')
        PC = np.pad(PC, (0, extra), mode='edge')
    
    return TR, PC

def grid_TR_sampling(n_points, TRs):
    npcs = math.ceil(n_points / len(TRs))
    PCs = np.linspace(0, 2*np.pi, npcs, endpoint=False)
    TR, PC = np.meshgrid(TRs, PCs,  indexing='ij')
    TR, PC = TR.flatten(), PC.flatten()
    TR, PC = TR[:n_points], PC[:n_points]
    return TR, PC


import os
import gasp

config = {
    'gasp_path': os.path.join(os.path.dirname(gasp.__file__)),
    'project_path': os.path.join(os.path.dirname(gasp.__file__), '../')
}

def get_gasp_path():
    ''' Retrieved gasp module path '''
    return config['gasp_path']

def get_project_path():
    ''' Retrieves gasp root project path '''
    return config['project_path']